
/*
 *
 *	stat.c
 *
 *	display file info
 *
 *	usage:   stat file
 *	         stat file file file ...
 *
 *	Goran Larsson
 *
 */
 
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <time.h>

#define MAXPATH 128

struct passwd *getpwuid ();
struct group  *getgrgid ();

main ( argc, argv )
int argc;
char *argv[];

{
	extern putmode ();
	struct stat buf;
	char cwd[MAXPATH];
	char *fyle;
	int i;

	if ( argc < 2 ) {
		printf ( "Usage: %s file..\n", argv[0] );
		exit ( 1 );
	}

	printf ( "\n" );

	for ( i = 1; i < argc; i++ ) {

		if ( i >= 2 ) {
			printf ( "----------------------------\n\n" );
		}

	/*	getcwd ( cwd, MAXPATH );	MISSING FROM libc.a !!!! */
		strcpy ( cwd, "" );

		fyle = argv[i];
		if ( fyle[0] == '/' ) {
			printf ( "File: %s\n", fyle );
		} else {
			printf ( "File: %s%s\n", cwd, fyle );
		}

		if ( stat ( fyle, &buf )) {
			printf ( "Cannot stat %s\n", fyle );
			continue;
		}

		printf ( "File mode          :  %o \t", buf.st_mode );
		putmode ( buf.st_mode );
		printf ( "\n" );

		printf ( "Inode number       :  %d\n", buf.st_ino );

		printf ( "Devide ID          :  %d \t%d,%d\n",
			buf.st_dev,
			( buf.st_dev >> 8 ) & 0xFF,
			buf.st_dev & 0xFF
		);

		printf ( "Special devide ID  :  %d \t%d,%d\n",
			buf.st_rdev,
			( buf.st_rdev >> 8 ) & 0xFF,
			buf.st_rdev & 0xFF
		);

		printf ( "Number of links    :  %d\n", buf.st_nlink );

		if ( getpwuid ( buf.st_uid ) == 0 ) {
			printf ( "User ID            :  %d \t(%s)\n",
				buf.st_uid,
				"******"
			);
		} else {
			printf ( "User ID            :  %d \t(%s)\n",
				buf.st_uid,
				getpwuid ( buf.st_uid ) -> pw_name
			);
		}

		if ( getgrgid ( buf.st_gid ) == 0 ) {
			printf ( "Group ID           :  %d \t(%s)\n",
				buf.st_gid,
				"******"
			);
		} else {
			printf ( "Group ID           :  %d \t(%s)\n",
				buf.st_gid,
				getgrgid ( buf.st_gid ) -> gr_name
			);
		}

		printf ( "Size in bytes      :  %d\n", buf.st_size );

		printf ( "Last access        :  %s",
			ctime ( &buf.st_atime )
		);

		printf ( "Last modification  :  %s",
			ctime ( &buf.st_mtime )
		);

		printf ( "Last status change :  %s",
			ctime ( &buf.st_ctime )
		);

		printf ( "\n" );
	}

	exit ( 0 );
}

putmode ( mode )
int mode;
{
	int i;

	for ( i = 2; i >= 0; i-- ) {
		switch ( i ) {
			case 2: printf ( "u:" ); break;
			case 1: printf ( "g:" ); break;
			case 0: printf ( "o:" ); break;
		}
		switch ( ( mode >> ( i * 3 ) ) & 0x7 ) {
			case 000 : printf ( "--- " ); break;
			case 001 : printf ( "--x " ); break;
			case 002 : printf ( "-w- " ); break;
			case 003 : printf ( "-wx " ); break;
			case 004 : printf ( "r-- " ); break;
			case 005 : printf ( "r-x " ); break;
			case 006 : printf ( "rw- " ); break;
			case 007 : printf ( "rwx " ); break;
		}
	}

	switch ( ( mode >> 12 ) & 0xF ) {
		case 010 : printf ( "       " ); break;
		case 004 : printf ( "dir    " ); break;
		case 002 : printf ( "char   " ); break;
		case 006 : printf ( "block  " ); break;
		case 007 : printf ( "xblock " ); break;
		case 003 : printf ( "xchar  " ); break;
		case 005 : printf ( "name   " ); break;
		case 001 : printf ( "pipe   " ); break;
		default:   printf ( "****** " ); break;
	}

	switch ( ( mode >> 10 ) & 0x3 ) {
		case 000 : printf ( "        " ); break;
		case 001 : printf ( "GID     " ); break;
		case 002 : printf ( "UID     " ); break;
		case 003 : printf ( "GID UID " ); break;
	}

	switch ( ( mode >> 9 ) & 0x1 ) {
		case 000 : printf ( "      " ); break;
		case 001 : printf ( "Stick " ); break;
	}
}
